/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "jsonStream.hh"
#include "TSeries.hh"
#include "fSeries/ASD.hh"
#include "fSeries/CSD.hh"
#include "fSeries/DFT.hh"
#include "fSeries/PSD.hh"
#include "DVecType.hh"
#include "json/writer.h"
#include <iostream>

using namespace std;
using namespace containers;

static char const * const tseries_typename = "tseries";
static char const * const fseries_typename = "fseries";
static char const * const fseries_modename[] = {"empty", "full", "full-real",
						"folded", "folded-odd", 0};

//===================================  double to time converter
inline Time
gps_time(double t0) {
   return Time(long(t0)) + Interval(t0 - long(t0));
}

//===================================  Constructor
jsonStream::jsonStream(void) {
   stack.reset(new jsonStack);
}

//===================================  Constructor
jsonStream::jsonStream(std::istream& in) {
   stack.reset(new jsonStack);
   parse(in);
}

//===================================  Get the array length
size_t
jsonStream::arrayLength(void) const {
   if (!isArray()) return 0;
   return stack->length();
}

//===================================  get a DVector
jsonStream::dvector_ptr
jsonStream::getDVector(void) const {
   string dtype("double");
   stack->fetch_data("data-type", dtype);
   DVector::DVType type = DVector::dvtype_code(dtype);
   dble_vect dvd;
   size_t N = stack->fetch_data("data", dvd);
   dvector_ptr r;
   if (type == DVector::DVector::t_dcomplex) {
      r.reset(new DVectW(N/2, reinterpret_cast<const dComplex*>(dvd.data())));
   } else if (type == DVector::t_complex) {
      DVectW dvw(N/2, reinterpret_cast<const dComplex*>(dvd.data()));
      r.reset(dvw.convert(type));
   } else {
      DVectD dvect(N, dvd.data());
      r.reset(dvect.convert(type));
   }
   return r;
}


//===================================  get FSeries from json document
jsonStream::fseries_ptr
jsonStream::getfSeries(void) const {
   fseries_ptr r;
   if (!isfSeries()) return r;

   string name, mode, subtype;
   stack->fetch_data("name", name);
   stack->fetch_data("subtype", subtype);

   //------------------------------------  get the data storage mode. 
   stack->fetch_data("mode", mode);
   fSeries::DSMode dsmode(fSeries::kEmpty);
   for (int i=0; fseries_modename[i]; i++) {
      if (mode == fseries_modename[i]) {
	 dsmode = fSeries::DSMode(i);
	 break;
      }
   }
   if (dsmode == fSeries::kEmpty) {
      throw runtime_error("jsonStream::getfSeries: invalid storage mode");
   }

   //------------------------------------  get the frequency and time limits.
   double x0=0, dx=0, t0=0, dt=0;
   stack->fetch_data("x0", x0);
   stack->fetch_data("dx", dx);
   stack->fetch_data("t0", t0);
   Time start = gps_time(t0);
   stack->fetch_data("duration", dt);
   Interval step(dt);
   dvector_ptr dvect(getDVector());

   //------------------------------------  set the
   fSeries fs(x0, dx, start, step, dvect.release());
   if (subtype == "fSeries") {
      r.reset(new containers::fSeries(fs));
   }
   else if (subtype == "asd") {
      r.reset(new containers::ASD);
      static_cast<fSeries&>(*r) = fs;
   }
   else if (subtype == "csd") {
      r.reset(new containers::CSD);
      static_cast<fSeries&>(*r) = fs;
   }
   else if (subtype == "dft") {
      r.reset(new containers::DFT);
      static_cast<fSeries&>(*r) = fs;
   }
   else if (subtype == "psd") {
      r.reset(new containers::PSD);
      static_cast<fSeries&>(*r) = fs;
   }
   else {
      cerr << "jsonStream::getfSeries: Unrecognized fseries type" << endl;
   }
   r->setName(name);
   return r;
}

//===================================  Build freq series from indexed element
jsonStream::fseries_ptr
jsonStream::getfSeries(size_t index) {
   if (isArray() && index < arrayLength()) {
      if (stack->push_element(index)) {
	 jsonStack::autopop pop_el(*stack);
	 return getfSeries();
      }
   }
   throw runtime_error("jsonStream: Unable to fetch indexed frequency series");
}

//===================================  build requested series from document
jsonStream::fseries_ptr
jsonStream::getfSeries(const std::string& req_name) {
   if (select(fseries_typename, req_name)) {
      jsonStack::autopop pop_el(*stack);
      return getfSeries();
   } else {
      return getfSeries();
   }
}

//===================================  Read fSeries from stream
jsonStream::fseries_ptr
jsonStream::getfSeries(std::istream& in, const std::string& name) {
   stack->clear();
   parse(in);
   return getfSeries(name);
}

//===================================  Build TSeries from top of stack
jsonStream::tseries_ptr
jsonStream::getTSeries(void) const {
   tseries_ptr r;
   if (!isTSeries()) return r;

   //------------------------------------  get the time series data.
   string name, units;
   double x0=0, dx=0;
   stack->fetch_data("name", name);
   stack->fetch_data("x0", x0);
   Time t0 = gps_time(x0);
   stack->fetch_data("dx", dx);
   Interval dt(dx);
   stack->fetch_data("units", units);

   //------------------------------------  set the pointer to a new time_series.
   r.reset(new TSeries(t0, dt, getDVector().release()));
   r->setName(name.c_str());
   r->setUnits(units);
   return r;
}

//===================================  Build time series from indexed element
jsonStream::tseries_ptr
jsonStream::getTSeries(size_t index) {
   if (isArray() && index < arrayLength()) {
      if (stack->push_element(index)) {
	 jsonStack::autopop pop_el(*stack);
	 return getTSeries();
      }
   }
   throw runtime_error("jsonStream: Unable to fetch indexed time series");
}

//===================================  Build time series from json document
jsonStream::tseries_ptr
jsonStream::getTSeries(const std::string& name) {
   if (select(tseries_typename, name)) {
      jsonStack::autopop pop_el(*stack);
      return getTSeries();
   } else {
      return getTSeries();
   }
}

//===================================  Read time series from stream
jsonStream::tseries_ptr
jsonStream::getTSeries(std::istream& in, const std::string& name) {
   stack->clear();
   parse(in);
   return getTSeries(name);
}

//===================================  Test for TSeries
bool
jsonStream::isTSeries(void) const {
   if (empty() || !stack->isStruct()) return false;
   string type;
   stack->fetch_data("type", type);
   return (type == tseries_typename);
}
 
//===================================  Constructor
bool
jsonStream::isfSeries(void) const {
   if (empty() || !stack->isStruct()) return false;
   string type;
   stack->fetch_data("type", type);
   return (type == fseries_typename);
}

//===================================  List type ans name of all objects.
void
jsonStream::listObjects(string_vect& names, string_vect& types) const {
}

//===================================  Constructor
void
jsonStream::parse(std::istream& in) {
   stack->clear();
   stack->parse(in);
}

//===================================  Push the specified structure.
bool
jsonStream::select(const std::string& type, const std::string& name) {
   string name_i, type_i;

   if (isArray()) {
      //----------------------------------  Array with no name requested
      if (name.empty()) {
	 if (stack->push_element(0)) {
	    stack->fetch_data("type", type_i);
	    if (type == type_i) return true;
	 }
	 cerr << "jsonStream::select: Unable to push element[0]" << endl;
      }
    
      //----------------------------------  Search array for type and name match
      else {
	 stack->reset_iterator();
	 while (stack->iterate()) {
	    stack->fetch_data("name", name_i);
	    stack->fetch_data("type", type_i);
	    if (name == name_i && type == type_i) return true;
	    stack->pop();
	 }
	 cerr << "jsonStream::select name (" << name << ") not found in array"
	      << endl; 
      }
   }
   //-----------------------------------  Single structure check the name & type
   else {
      stack->fetch_data("name", name_i);
      stack->fetch_data("type", type_i);
      if ((name.empty() || name == name_i) && type == type_i) return false;
   }
   throw runtime_error("jsonStream::select requested series not in stream.");
}

//===================================  Get fSeries Name
inline string
fs_type(fSeries::FSType t) {
   switch (t) {
   case fSeries::kSeries:
      return "fSeries";
      break;
   case fSeries::kDFT:
      return "DFT";
      break;
   case fSeries::kPSD:
      return "PSD";
      break;
   case fSeries::kCSD:
      return "CSD";
      break;
   case fSeries::kASD:
      return "ASD";
      break;
   }
   throw runtime_error("jsonStream: Unidentified fSeries subtype"); 
}

//===================================  Add DVector data to a document
void
jsonStream::putDVector(Json::Value& doc, const DVector& dv) const {
   doc["data-type"] = string(dv.getTypeName());
   size_t N = dv.size();
   if (dv.C_data()) {
      const DVectC& dvc(dynamic_cast<const DVectC&>(dv));
      for (size_t i=0; i<N; i++) {
	 doc["data"][Json::ArrayIndex(2*i)]   = double(dvc[i].Real());
	 doc["data"][Json::ArrayIndex(2*i+1)] = double(dvc[i].Imag());
      }
   } else if (dv.W_data()) {
      const DVectW& dvw(dynamic_cast<const DVectW&>(dv));
      for (size_t i=0; i<N; i++) {
	 doc["data"][Json::ArrayIndex(2*i)]   = dvw[i].Real();
	 doc["data"][Json::ArrayIndex(2*i+1)] = dvw[i].Imag();
      }
   } else {
      DVectD dvd(dv);
      for (size_t i=0; i<N; i++) {
	 doc["data"][Json::ArrayIndex(i)] = dvd[i];
      }
   }
}

//===================================  Write out the structure contents
void
jsonStream::putfSeries(const containers::fSeries& fs, std::ostream& out) {
    Json::Value  doc;
    doc["type"]    = fseries_typename;
    doc["name"]    = fs.getName();
    fSeries::FSType ft = fs.getFSType();
    doc["subtype"] = fs_type(ft);
    doc["x0"]      = fs.getLowFreq();
    doc["dx"]      = fs.getFStep();
    doc["t0"]      = fs.getStartTime().totalS();
    doc["dt"]      = double(fs.getDt());
    putDVector(doc, fs.refDVect());
    out << doc;
}
 
//===================================  Push the specified structure.
void 
jsonStream::putTSeries(const TSeries& ts, std::ostream& out) {
   Json::Value  doc;
   doc["type"]  = tseries_typename;
   doc["name"]  = ts.getName();
   doc["x0"]    = ts.getStartTime().totalS();
   doc["dx"]    = double(ts.getTStep());
   doc["units"] = string(ts.getUnits());
   putDVector(doc, *ts.refDVect());
   out << doc;
}
